/* Copyright (C) 2000 drscholl@users.sourceforge.net
   This is free software distributed under the terms of the
   GNU Public License.  See the file COPYING for details.

   $Id: main.c,v 1.237 2000/10/30 23:42:47 drscholl Exp $ */

#ifdef WIN32
#include <windows.h>
#include <winsock.h>
#endif /* WIN32 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>
#endif /* !WIN32 */
#if HAVE_LIBWRAP
#include <tcpd.h>
#endif
#include "opennap.h"
#include "debug.h"
#include "libpoll.h"

#define ON_POLL_ACCEPT	(void*)0x1
#define ON_POLL_STATS	(void*)0x2

/*
 * Global Variables
 */

LIST   *Bans = 0;		/* bans on ip addresses / users */
char    Buf[2048];		/* global scratch buffer */

double  Bytes_In = 0;
double  Bytes_Out = 0;
HASH   *Channel_Db;
int     Client_Queue_Length;
HASH   *Channels;		/* global channel list */

#if VERSION_STATS
HASH   *Client_Versions;
#endif
int     Compression_Level = 0;
char   *Config_Dir = SHAREDIR;
u_int   Connection_Count = 0;
time_t  Current_Time = 0;
LIST   *Destroy = 0;
int     Flood_Commands;
int     Flood_Time;
LIST   *Flooders = 0;
HASH   *Hotlist;		/* global hotlist */
u_int   Interface = INADDR_ANY;
time_t  Last_Click = 0;
char   *Listen_Addr = 0;
int     Login_Timeout;
int     Max_Browse_Result;
int     Max_Client_String;
int     Max_Clones;
int     Max_Command_Length;
int     Max_Connections;
int     Max_Hotlist;
int     Max_Ignore;
int     Max_Reason;
int     Max_Search_Results;
int     Max_Time_Delta;
int     Max_Topic;
int     Max_User_Channels;	/* default, can be changed in config */
int     Nick_Expire;
double  Num_Gigs = 0;		/* in kB */
int     Num_Files = 0;
int     Ping_Interval;
lp_ctx *Poll_Context;
int     Search_Timeout;
char   *Server_Alias = 0;
LIST   *Server_Auth = 0;
u_int   Server_Ip = 0;
u_int   Server_Flags = 0;
char   *Server_Name = 0;
LIST   *Server_Ports = 0;	/* which port(s) to listen on for connections */
int     Server_Queue_Length;
int     SigCaught = 0;
int     Stat_Click;		/* interval (in seconds) to send server stats */
time_t  Server_Start;		/* time at which the server was started */
u_int   Total_Bytes_In = 0;	/* bytes received */
u_int   Total_Bytes_Out = 0;	/* bytes sent */
int     User_Db_Interval;	/* how often to save the user database */
HASH   *Users;			/* global users list */
int     Warn_Time_Delta;

#ifndef ROUTING_ONLY
int     File_Count_Threshold;
HASH   *File_Table;		/* global file list */
int     Local_Files = 0;	/* number of files shared by local users */
int     Max_Shared;

#if RESUME
HASH   *MD5;			/* global hash list */
#endif /* RESUME */
int     Stats_Port;		/* port to listen on for stats info */
#endif /* ! ROUTING_ONLY */

#ifndef WIN32
int     Uid;
int     Gid;
int     Connection_Hard_Limit;
int     Max_Data_Size;
int     Max_Rss_Size;
#endif
int     Max_Nick_Length;
int     Max_Channel_Length = 0;

/* local clients (can be users or servers) */
CONNECTION **Clients = NULL;
int     Num_Clients = 0;
int     Max_Clients = 0;

/* local server list.  NOTE that this contains pointers into the Clients
   list to speed up server-server message passing */
LIST   *Servers = 0;

/* list of all servers in the cluster */
LIST   *Server_Links = 0;

/* Cache of server names for caching the user->server pointers */
LIST   *Server_Names = 0;

#define BACKLOG 50

static void
update_stats (void)
{
    int     i, l;
    int     numServers = list_count (Servers);
    time_t  delta;

    delta = Current_Time - Last_Click;

    strcpy (Buf, ctime (&Server_Start));
    Buf[strlen (Buf) - 1] = 0;
    log ("update_stats: server was started on %s", Buf);
    strcpy (Buf, ctime (&Current_Time));
    Buf[strlen (Buf) - 1] = 0;
    log ("update_stats: current time is %s", Buf);
    log ("update_stats: library is %.0f GB, %d files, %d users",
	 Num_Gigs / 1048576., Num_Files, Users->dbsize);
    log ("update_stats: %d local clients, %d linked servers",
	 Num_Clients - numServers, numServers);
#ifndef ROUTING_ONLY
    log ("update_stats: %d local files", Local_Files);
    log ("update_stats: File_Table contains %d entries", File_Table->dbsize);
#endif
    log ("update_stats: %.0f searches/sec",
	 (float)Search_Count / (float)delta);
    log ("update_stats: User_Db contains %d entries", User_Db->dbsize);
    log ("update_stats: %d channels", Channels->dbsize);
    log ("update_stats: %.2f kbytes/sec in, %.2f kbytes/sec out",
	 Bytes_In / 1024. / delta, Bytes_Out / 1024. / delta);
    Total_Bytes_In += Bytes_In;
    Total_Bytes_Out += Bytes_Out;
    log ("update_stats: %u bytes sent, %u bytes received",
	 Total_Bytes_Out, Total_Bytes_In);
#if HAVE_POLL && !defined(HAVE_KQUEUE)
    log ("update_stats: Poll_Context->num_fds=%d", Poll_Context->num_fds);
#endif

    /* reset counters */
    Bytes_In = 0;
    Bytes_Out = 0;
    Search_Count = 0;
    Last_Click = Current_Time;

    /* since we send the same data to many people, optimize by forming
       the message once then writing it out */
    snprintf (Buf + 4, sizeof (Buf) - 4, "%d %d %.0f", Users->dbsize,
	      Num_Files, Num_Gigs / 1048576.);
    set_tag (Buf, MSG_SERVER_STATS);
    l = strlen (Buf + 4);
    set_len (Buf, l);
    l += 4;
    for (i = 0; i < Num_Clients; i++)
    {
	if (ISUSER (Clients[i]))
	    queue_data (Clients[i], Buf, l);
    }
}

#if HAVE_LIBWRAP
int     allow_severity = 0;
int     deny_severity = 0;
int     hosts_ctl (char *, char *, char *, char *);
#endif

/* accept all pending connections */
static void
accept_connection (int s)
{
    CONNECTION *cli;
    socklen_t sinsize;
    struct sockaddr_in sin;
    int     f;

    for (;;)
    {
	sinsize = sizeof (sin);
	if ((f = accept (s, (struct sockaddr *)&sin, &sinsize)) < 0)
	{
	    if (N_ERRNO != EWOULDBLOCK)
		nlogerr ("accept_connection", "accept");
	    return;
	}
#if HAVE_LIBWRAP
	if (!hosts_ctl (PACKAGE, STRING_UNKNOWN, inet_ntoa (sin.sin_addr),
			STRING_UNKNOWN))
	{
	    log ("accept_connection: tcp wrappers denied %s",
		 inet_ntoa (sin.sin_addr));
	    CLOSE (f);
	    return;
	}
#endif
	if ((cli = new_connection ()) == 0)
	{
	    CLOSE (f);
	    return;
	}
	cli->fd = f;
	/* if we have a local connection, use the external
	   interface so others can download from them */
	if (sin.sin_addr.s_addr == inet_addr ("127.0.0.1"))
	{
	    log
		("accept_connection: connected via loopback, using external ip");
	    cli->ip = Server_Ip;
	    cli->host = STRDUP (Server_Name);
	    if (!cli->host)
	    {
		OUTOFMEMORY ("accept_connection");
		goto error;
	    }
	}
	else
	{
	    cli->ip = BSWAP32 (sin.sin_addr.s_addr);
	    cli->host = STRDUP (inet_ntoa (sin.sin_addr));
	    if (!cli->host)
	    {
		OUTOFMEMORY ("accept_connection");
		goto error;
	    }
	}
	cli->port = ntohs (sin.sin_port);
	cli->class = CLASS_UNKNOWN;
	cli->timer = Current_Time;	/* set a login timer */
	if (add_client (cli, 1))
	    return;
	set_nonblocking (f);
	set_keepalive (f, 1);	/* enable tcp keepalive messages */
    }
    /* not reached */
    ASSERT (0);
    return;
  error:
    CLOSE (f);
    if (cli->host)
	FREE (cli->host);
    FREE (cli);
}

#ifndef ROUTING_ONLY
static void
report_stats (int fd)
{
    int     n;
    struct sockaddr_in sin;
    socklen_t sinsize = sizeof (sin);
    float   loadavg = 0;

    n = accept (fd, (struct sockaddr *)&sin, &sinsize);
    if (n == -1)
    {
	nlogerr ("report_stats", "accept");
	return;
    }
    log ("report_stats: connection from %s:%d", inet_ntoa (sin.sin_addr),
	 htons (sin.sin_port));
#ifdef linux
    {
	FILE   *f = fopen ("/proc/loadavg", "r");

	if (f)
	{
	    fscanf (f, "%f", &loadavg);
	    fclose (f);
	}
	else
	{
	    log ("report_stats: /proc/loadavg: %s (errno %d)",
		 strerror (errno), errno);
	}
    }
#endif /* linux */
    snprintf (Buf, sizeof (Buf), "%d %d %.2f %.0f 0\n", Users->dbsize,
	      Num_Files, loadavg, Num_Gigs * 1024.);
    WRITE (n, Buf, strlen (Buf));
    CLOSE (n);
}
#endif

static void
usage (void)
{
    fprintf (stderr,
	     "usage: %s [ -bhsv ] [ -c DIR ] [ -p PORT ] [ -l IP ]\n",
	     PACKAGE);
    fprintf (stderr, "  -c DIR	read config files from DIR (default: %s)\n",
	     SHAREDIR);
    fputs ("  -b		run as a background process (daemon)\n",
	   stderr);
    fputs ("  -h		print this help message\n", stderr);
    fputs
	("  -l IP		listen only on IP instead of all interfaces\n",
	 stderr);
    fputs ("  -p PORT	listen on PORT for connections (default: 8888)\n",
	   stderr);
    fputs
	("  -s		channels may only be created by privileged users\n",
	 stderr);
    fputs ("  -v		display version information\n", stderr);
    exit (0);
}

static void
version (void)
{
    fprintf (stderr, "%s %s\n", PACKAGE, VERSION);
    fprintf (stderr, "Copyright (C) 2000 drscholl@users.sourceforge.net\n");
    exit (0);
}

static int *
args (int argc, char **argv, int *sockfdcount)
{
    int     i;
    LIST   *ports = 0, *tmpList;
    int     iface = -1;
    int    *sockfd;
    int     not_root = 1;
    int     port;

#ifndef WIN32
    not_root = (getuid () != 0);
#endif

    while ((i = getopt (argc, argv, "bc:hl:p:svD")) != -1)
    {
	switch (i)
	{
	case 'b':
	    Server_Flags |= ON_BACKGROUND;
	    break;
	case 'D':
	    Server_Flags |= ON_NO_LISTEN;	/* dont listen on stats port */
	    break;
	case 'c':
	    /* ignore the command line option if we're running as root.
	     * we don't allow non-root users to specify their own config
	     * files to avoid possible security problems.
	     */
	    if (not_root)
		Config_Dir = optarg;
	    else
	    {
		log ("args: can't use -c when run as root");
		exit (1);
	    }
	    break;
	case 'l':
	    iface = inet_addr (optarg);
	    break;
	case 'p':
	    /* don't allow a privileged port to be used from the command line
	     * if running as root.  this can only be specified in the
	     * configuration file defined at compile time.
	     */
	    port = atoi (optarg);
	    if (not_root || port > 1024)
	    {
		tmpList = CALLOC (1, sizeof (LIST));
		tmpList->data = STRDUP (optarg);
		tmpList->next = ports;
		ports = tmpList;
	    }
	    else
	    {
		log ("args: privileged ports not allowed on command line");
		exit (1);
	    }
	    break;
	case 's':
	    Server_Flags |= ON_STRICT_CHANNELS;
	    break;
	case 'v':
	    version ();
	    break;
	default:
	    usage ();
	}
    }

#if USE_CHROOT
    /* we always use the compiled directory instead of the one on the command
     * line here to avoid problems.
     */
    if (chroot (SHAREDIR))
    {
	perror ("chroot");
	exit (1);
    }
    if (chdir ("/"))
    {
	perror ("chdir");
	exit (1);
    }
    /* force the config files to be relative to the chroot jail */
    Config_Dir = "/";
    /* privs will be dropped later.  we still need them to read the the
     * config file and set resources.
     */
#endif

#if !defined(WIN32) && !defined(__EMX__)
    /* check whether to run in the background */
    if (Server_Flags & ON_BACKGROUND)
    {
	if (fork () == 0)
	    setsid ();
	else
	    exit (0);
    }
#endif

    if (init_server ())
	exit (1);

    /* if the interface was specified on the command line, override the
     * value from the config file
     */
    if (iface != -1)
    {
	Interface = iface;
	Server_Ip = iface;
    }

    /* if port(s) were specified on the command line, override the values
       specified in the config file */
    if (!ports)
	ports = Server_Ports;

    /* create the incoming connections socket(s) */
    *sockfdcount = list_count (ports);
    /* ensure at least one valid port */
    if (*sockfdcount < 1)
    {
	log ("args: no server ports defined");
	exit (1);
    }
    sockfd = CALLOC (*sockfdcount, sizeof (int));

    log ("args: listening on %d sockets", *sockfdcount);
    for (i = 0, tmpList = ports; i < *sockfdcount;
	 i++, tmpList = tmpList->next)
    {
	if ((sockfd[i] = new_tcp_socket (ON_NONBLOCKING | ON_REUSEADDR)) < 0)
	    exit (1);
	if (bind_interface (sockfd[i], Interface, atoi (tmpList->data)) == -1)
	    exit (1);
	if (listen (sockfd[i], BACKLOG) < 0)
	{
	    nlogerr ("args", "listen");
	    exit (1);
	}
	log ("args: listening on %s port %d", my_ntoa (Interface),
	     atoi (tmpList->data));
    }
    if (ports != Server_Ports)
	list_free (ports, free_pointer);
    return sockfd;
}

/* sync in-memory state to disk so we can restore properly */
static void
dump_state (void)
{
    userdb_dump ();		/* write out the user database */
#ifndef ROUTING_ONLY
    save_bans ();		/* write out server bans */
    dump_channels ();		/* write out persistent channels file */
#endif
}

#ifndef ROUTING_ONLY
static int
init_stats_port (void)
{
    int     sp = -1;

    if (!option (ON_NO_LISTEN) && Stats_Port != -1)
    {
	/* listen on port 8889 for stats reporting */
	if ((sp = new_tcp_socket (ON_REUSEADDR)) == -1)
	    exit (1);
	if (bind_interface (sp, Interface, Stats_Port))
	    exit (1);
	if (listen (sp, BACKLOG))
	{
	    logerr ("main", "listen");
	    exit (1);
	}
    }
    return sp;
}
#endif

/* puts the specified connection on the destroy list to be reaped at the
 * end of the main event loop
 */
void
destroy_connection (CONNECTION * con)
{
    LIST   *list;

    if (con->destroy)
    {
	list = list_find (Destroy, con);
	if (list)
	    return;		/* already destroyed */
	log
	    ("destroy_connection: error, destroyed connection not on Destroy list");
	log ("destroy_connection: con->host = %s", con->host);
	if (ISUSER (con))
	    log ("destroy_connection: con->user->nick = %s", con->user->nick);
    }

    /* append to the list of connections to destroy */
    list = CALLOC (1, sizeof (LIST));
    if (!list)
    {
	OUTOFMEMORY ("destroy_connection");
	return;
    }
    list->data = con;
    Destroy = list_push (Destroy, list);
    send_queued_data (con);	/* force a flush */
    con->destroy = 1;
#if DEBUG
    ASSERT (lp_flags (Poll_Context, con->fd) != 0);
#endif
    /* we don't want to read/write anything furthur to this fd */
    lp_clear (Poll_Context, con->fd, LP_RDWR);
}

static void
reap_dead_connection (CONNECTION * con)
{
    ASSERT (validate_connection (con));
    if (con->id < Num_Clients - 1)
    {
	/* swap this place for the last connection in the array */
	Clients[Num_Clients - 1]->id = con->id;
	Clients[con->id] = Clients[Num_Clients - 1];
    }
    Num_Clients--;
    remove_connection (con);
}

int
add_client (CONNECTION * cli, int inbound)
{
    if (Max_Clients == Num_Clients)
    {
	/* no space left, allocate more */
	if (safe_realloc
	    ((void **)&Clients, sizeof (CONNECTION *) * (Max_Clients + 10)))
	{
	    OUTOFMEMORY ("add_client");
	    CLOSE (cli->fd);
	    FREE (cli->host);
	    FREE (cli);
	    return -1;
	}
	Max_Clients += 10;
    }
    Clients[Num_Clients] = cli;
    cli->id = Num_Clients;
    Num_Clients++;
    lp_set (Poll_Context, cli->fd, inbound ? LP_READ : LP_WRITE, cli);
    return 0;
}

static void
handle_io (int fd, int flags, void *fd_data, void *cb_data)
{
    (void)cb_data;

    if (fd_data == ON_POLL_ACCEPT)
    {
	if (flags & LP_ERROR)
	{
	    SigCaught = 1;
	    log ("handle_io: error on listen socket!");
	    return;
	}
	accept_connection (fd);
    }
#ifndef ROUTING_ONLY
    else if (fd_data == ON_POLL_STATS)
    {
	if (flags & LP_ERROR)
	{
	    /* stop reading to prevent future errors */
	    lp_clear (Poll_Context, fd, LP_READ);
	    log ("handle_io: error on stats socket!");
	    return;
	}
	report_stats (fd);
    }
#endif
    else
    {
	CONNECTION *con = (CONNECTION *) fd_data;

	/* not sure why this occasionally happens, but catch it to prevent
	 * coring the server.  possibly a bug in either kqueue() or in
	 * libpoll's interface to kqueue() ???
	 */
	if (!con)
	{
	    log ("handle_io: missing context for fd %d\n", fd);
	    return;
	}

	ASSERT (validate_connection (con));
	if (con->fd != fd)
	{
	    log ("handle_io: error, fd (%d) does not match context (%d)",
		 fd, con->fd);
	}
	if (flags & LP_ERROR)
	{
	    socklen_t len = sizeof (flags);

	    if (getsockopt (con->fd, SOL_SOCKET, SO_ERROR, SOCKOPTCAST &flags,
			&len))
	    {
		logerr ("handle_io", "getsockopt");
	    }
	    else
	    {
		log ("handle_io: %s (fd %d): %s (errno %d)", con->host,
		     con->fd, strerror (N_ERRNO), N_ERRNO);
	    }
	    destroy_connection (con);
	    return;
	}
	if (flags & LP_READ)
	{
	    if (!con->destroy)
	    {
		handle_connection (con);
	    }
	}
	if (flags & LP_WRITE)
	{
	    if (con->connecting)
		complete_connect (con);
	    else if (send_queued_data (con) == -1)
		destroy_connection (con);
	}
    }
}

static void
flood_expire (void)
{
    LIST  **list, *tmp;
    CONNECTION *con;

    for (list = &Flooders; *list;)
    {
	con = (*list)->data;
	if (con->flood_start + Flood_Time < Current_Time)
	{
	    /* flood timer expired, resume reading commands */
	    lp_set (Poll_Context, con->fd, LP_READ, con);
	    tmp = *list;
	    *list = (*list)->next;
	    FREE (tmp);
	}
	else
	    list = &(*list)->next;
    }
}

/* since server->server data is always queued up so it can be compressed
 * in one shot, we have to explicitly call send_queued_data() for each
 * server here.
 */
static void
flush_server_data (CONNECTION * con)
{
    ASSERT (validate_connection (con));
    if (send_queued_data (con) == -1)
	destroy_connection (con);
}

/* reap connections which have no logged in after `Login_Timeout' seconds
 */
static void
login_timeout (void)
{
    int i;

    for (i = 0; i < Num_Clients; i++)
    {
	if (ISUNKNOWN(Clients[i]) &&
		Current_Time - Clients[i]->timer > Login_Timeout)
	{
	    log("login_timeout: terminating %s", Clients[i]->host);
	    destroy_connection(Clients[i]);
	}
    }
}

int
main (int argc, char **argv)
{
    int    *sockfd;		/* server sockets */
    int     sockfdcount;	/* number of server sockets */
    int     sp = -1;		/* stats port */
    int     i;			/* generic counter */
    int     timeout;

#ifdef WIN32
    WSADATA wsa;

    WSAStartup (MAKEWORD (1, 1), &wsa);
#endif /* !WIN32 */


    /* minimize the stack space for the main loop by moving the command line
       parsing code to a separate routine */
    sockfd = args (argc, argv, &sockfdcount);

    Poll_Context = lp_init ();
    for (i = 0; i < sockfdcount; i++)
	lp_set (Poll_Context, sockfd[i], LP_READ, ON_POLL_ACCEPT);

#ifndef ROUTING_ONLY
    sp = init_stats_port ();
    if (sp != -1)
	lp_set (Poll_Context, sp, LP_READ, ON_POLL_STATS);
#endif

    /* schedule periodic events */
    add_timer (Stat_Click, -1, (timer_cb_t) update_stats, 0);
    add_timer (User_Db_Interval, -1, (timer_cb_t) dump_state, 0);
    add_timer (60, -1, (timer_cb_t) expire_bans, 0);
    add_timer (Ping_Interval, -1, (timer_cb_t) lag_detect, 0);
    add_timer (Login_Timeout, -1, (timer_cb_t) login_timeout, 0);

    /* initialize so we get the correct delta for the first call to
       update_stats() */
    Last_Click = Current_Time;

    /* main event loop */
    while (!SigCaught)
    {
	Current_Time = time (0);

	timeout = next_timer ();
	/* if we have a flood list and the timeout is greater than when
	 * the flood expires, reset the timeout
	 */
	if (Flooders && timeout > Flood_Time)
	    timeout = Flood_Time;

	if (lp_poll (Poll_Context, timeout * 1000) < 0)
	{
	    continue;
	}

	/* this is where the real work gets done.  handle_io() gets called
	 * when one of the client connections becomes readable/writable
	 */
	lp_dispatch (Poll_Context, handle_io, 0);

	flood_expire ();

	/* remove destroyed connections from the client list */
	list_free (Destroy, (list_destroy_t) reap_dead_connection);
	Destroy = 0;

	list_foreach (Servers, (list_callback_t) flush_server_data);

	/* execute any pending events now */
	exec_timers (Current_Time);
    }

    log ("main: shutting down");

    /* disallow incoming connections */
    for (i = 0; i < sockfdcount; i++)
	CLOSE (sockfd[i]);
#ifndef ROUTING_ONLY
    /* disallow stats reporting */
    if (sp != -1)
	CLOSE (sp);
#endif

    dump_state ();		/* save to disk */

    /* close all client connections */
    for (i = 0; i < Num_Clients; i++)
    {
#if DEBUG
	remove_connection (Clients[i]);
#else
	/* just close the fd for the quickest exit possible.  we
	 * don't need to shut down cleanly since our state is
	 * already saved to disk
	 */
	CLOSE (Clients[i]->fd);
#endif
    }

    /* only clean up memory if we are in debug mode, its kind of pointless
       otherwise */
#if DEBUG
    lp_final (Poll_Context);

    motd_close ();

    if (sockfd)
	FREE (sockfd);

    /* clean up */
    if (Clients)
	FREE (Clients);

    if (Servers)
	FREE (Servers);

#ifndef ROUTING_ONLY
    free_hash (File_Table);
#if RESUME
    free_hash (MD5);
#endif /* RESUME */
#if VERSION_STATS
    free_hash (Client_Versions);
#endif
    free_hash (Filter);
#endif

    free_hash (Users);
    free_hash (Channels);
    free_hash (Hotlist);
    free_hash (User_Db);
    free_timers ();

    list_free (Bans, (list_destroy_t) free_ban);
    list_free (Server_Auth, (list_destroy_t) free_server_auth);
    list_free (Server_Names, free_pointer);

    /* free up memory associated with global configuration variables */
    free_config ();

    /* this displays a list of leaked memory.  pay attention to this. */
    CLEANUP ();
#endif

#ifdef WIN32
    WSACleanup ();
#endif

    Current_Time = time (0);
    log ("main: server ended at %s", ctime (&Current_Time));

    exit (0);
}
